/////////////////////////////////////////////////////////////////////////////
// Name:        Archive.cpp
// Purpose:    Non-virtual archive stuff
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     GPL v3
/////////////////////////////////////////////////////////////////////////////
// For compilers that support precompilation, includes "wx/wx.h".  
#include "wx/wxprec.h" 

#ifndef WX_PRECOMP
  #include "wx/wx.h"
  #include "wx/xrc/xmlres.h"
  #include "wx/config.h"
#endif


#include "MyGenericDirCtrl.h"
#include "Externs.h"
#include "ExecuteInDialog.h"
#include "Misc.h"
#include "Tools.h"
#include "Filetypes.h"
#include "MyDirs.h"
#include "MyFrame.h"
#include "Archive.h"
#include "wx/tarstrm.h"




void MyListBox::Add( wxString newfile )  // Appends to the list only if not already present
{
if ( newfile.IsEmpty() ) return;

if ( FindString( newfile ) == wxNOT_FOUND )
  Append( newfile );
}

void MyListBox::OnKey(wxKeyEvent& event)  // If Delete key pressed while file(s) highlit, delete them
{
if ( event.GetKeyCode() == WXK_DELETE )                     // If Del pressed
  {  wxArrayInt selections;
    int count = GetSelections(selections);                  // Get selected files into array
    if ( ! count ) return;
    
    for ( int n=count; n > 0; --n )
      Delete( selections[n-1] );
  }
 else 
  event.Skip();
}

IMPLEMENT_DYNAMIC_CLASS(MyListBox,wxListBox)

BEGIN_EVENT_TABLE(MyListBox,wxListBox)
  EVT_CHAR(MyListBox::OnKey)
END_EVENT_TABLE()
//-----------------------------------------------------------------------------------------------------------------------
BEGIN_EVENT_TABLE(DecompressOrExtractDlg,wxDialog)
EVT_BUTTON(XRCID("Compressed"),DecompressOrExtractDlg::OnDecompressButton)
EVT_BUTTON(XRCID("Archive"),DecompressOrExtractDlg::OnExtractButton)
END_EVENT_TABLE()
//-----------------------------------------------------------------------------------------------------------------------

void ArchiveDialogBase::init( Archive* dad )
{
parent = dad;

combo = (wxComboBox*)FindWindow( wxT("Combo") );
list = (MyListBox*)FindWindow( wxT("CurrentFiles") );

parent->FilepathArray.Clear();                              // Empty the array first: some paths to this method will have called GetMultiplePaths() already
parent->active->GetMultiplePaths( parent->FilepathArray );  // Load the selected files into the stringarray

      // The next line means that a <CR> in combo's textctrl is equivalent to pressing the AddFile button
Connect( XRCID("Combo"), wxEVT_COMMAND_TEXT_ENTER, (wxObjectEventFunction)&ArchiveDialogBase::OnAddFile );
}

//static
wxString ArchiveDialogBase::FindUncompressedName( wxString& archivename, enum ziptype ztype )  // What'd a compressed archive be called if uncompressed?
{
wxString UncompressedName; int c;

if ( archivename.IsEmpty() || ztype==taronly ) return wxEmptyString;

if ( (c = archivename.Find( wxT(".tar.") )) != -1 )         // If this is a tar.?z(2) archive
  UncompressedName = archivename.Left( c + 4 );             //   remove the .?z(2), keeping the .tar
 else                                                       // Otherwise it must have a .t?z ext
   UncompressedName = archivename.BeforeLast( wxT('.') ) + wxT(".tar");  // Replace this with tar

return UncompressedName;
}

//static
bool ArchiveDialogBase::IsAnArchive( wxString filename )    // Decide if the file is an archive suitable for appending to
{
wxString ext = filename.AfterLast('.'); ext.MakeLower();    // See if the last .ext is relevant
if ( ext==wxT("tar") || ext==wxT("tgz") || ext==wxT("tbz2") || ext==wxT("tbz") || ext==wxT("zip") || ext==wxT("htb") )
  return true;
  
return ( filename.Contains( wxT(".tar.") ) );               // Otherwise see if there's a less-terminal tar ext
}

void ArchiveDialogBase::OnAddFile(wxCommandEvent& WXUNUSED(event))  // Add the contents of combo to the listbox
{
for ( size_t n=0; n < combo->GetCount(); ++n )              // First go thru the combobox proper
  list->Add( combo->GetString( n ) );                       // 'Add' checks for duplicates
  
wxString newfile = combo->GetValue();                       // Now deal with the textctrl bit, in case an entry was written in
if ( ! newfile.IsEmpty() )
  {  wxString tempfile = newfile;
    if ( newfile.Left(2) == wxT("./") )                      // If asking for the 'current' dir, get it from active pane
      tempfile = parent->active->GetActiveDirPath() + newfile.Mid(1);
     else
       if ( newfile.GetChar(0) != wxT('/') )                 // If relative address for this file, similarly
        { tempfile = parent->active->GetActiveDirPath() + wxT('/'); tempfile += newfile; }
    if ( ! ( wxFileExists( tempfile ) || wxDirExists( tempfile ) ) )  // Since this entry was user-inserted, not from Browsing, check for rubbish
      {  wxMessageDialog dialog(this, _("I can't seem to find this file or directory.\nTry using the Browse button."),  _("Oops!"), wxOK |wxICON_ERROR );
        dialog.ShowModal();
        combo->SetValue(wxT(""));
      }
     else  
        list->Add( newfile );
  }

combo->Clear(); combo->SetValue(wxT(""));
}

void ArchiveDialogBase::OnFileBrowse(wxCommandEvent& WXUNUSED(event))  // Browse for files to add to listbox
{
FileDirDlg fdlg( this, FD_MULTIPLE | FD_SHOWHIDDEN, parent->active->GetActiveDirPath(), _("Choose file(s) and/or directories") );
if ( fdlg.ShowModal() != wxID_OK ) return;

wxArrayString filenames;
fdlg.GetPaths( filenames );                                  // Get the selection(s) into an array

for ( size_t n=0; n < filenames.GetCount(); ++n )            // Store the answers in the combobox.  The AddFile button will pass them to the listbox
  if ( filenames[n].AfterLast(wxFILE_SEP_PATH) != wxT("..") && filenames[n].AfterLast(wxFILE_SEP_PATH) != wxT(".") )
      combo->Append( filenames[n] );
#if wxVERSION_NUMBER >= 2701 && (defined(__WXGTK20__) || defined(__WXX11__))  
if ( ! combo->IsEmpty() )  combo->SetSelection(0);           // In gtk2 and >2.7.0.1 we need explicitly to set the value
#endif
}

BEGIN_EVENT_TABLE(ArchiveDialogBase,wxDialog)
  EVT_BUTTON(XRCID("AddFile"), ArchiveDialogBase::OnAddFile)
  EVT_BUTTON(XRCID("Browse"), ArchiveDialogBase::OnFileBrowse)
END_EVENT_TABLE()
//-----------------------------------------------------------------------------------------------------------------------

void MakeArchiveDialog::Init( Archive* dad, bool newarch )
{
init( dad );                                                 // Do the base-class init first

NewArchive = newarch;
bool foundarchive = false;

createincombo = (wxComboBox*)FindWindow( wxT("CreateInCombo") );

                    // Set checkboxes to last-used values
((wxCheckBox*)FindWindow( wxT("DeReference") ))->SetValue( parent->DerefSymlinks );
((wxCheckBox*)FindWindow( wxT("Verify") ))->SetValue( parent->Verify );
((wxCheckBox*)FindWindow( wxT("DeleteSources") ))->SetValue( parent->DeleteSource );
if ( NewArchive )
    ((wxRadioBox*)FindWindow( wxT("CompressRadio") ))->SetSelection( parent->ArchiveCompress );

for ( size_t n=0; n < parent->ArchiveHistory.GetCount(); ++n ) // Load the archive path history into the combo
  createincombo->Append( parent->ArchiveHistory[n] );
if ( NewArchive )   createincombo->SetValue(wxT("./"));      // If creating, give archive-path a default value of ./
  else          createincombo->SetValue(wxT(""));            // Otherwise clear the textctrl bit

size_t count = parent->FilepathArray.GetCount();
for ( size_t n=0; n < count; ++n )                           //   & the listbox
  if ( ! (parent->FilepathArray[n]==parent->active->GetActiveDirectory() && count==1) )  // We don't want rootdir in the list just because it's automatically highlit
    {  if ( ! (NewArchive || foundarchive) )      // If we're appending, see if we can spot an archive name to append to (if we haven't found one already)
        {  if ( IsAnArchive( parent->FilepathArray[n] ) )    //  See if the file is an archive suitable for appending  to
            {  createincombo->SetValue(parent->FilepathArray[n]);  // If so, put it in the archive filepath combobox
              foundarchive = true;
            }
            else list->Append( parent->FilepathArray[n] );    // Otherwise put in the files-to-be-added listbox as usual
        }
       else list->Append( parent->FilepathArray[n] );         // Otherwise put in the files-to-be-added listbox
    }
}

void MakeArchiveDialog::OnCheckBox(wxCommandEvent& event)  // Make sure there's only one type of archiving checked at a time
{
int id = event.GetId();

if ( id == XRCID("UseTar") )
  ((wxCheckBox*)FindWindow( wxT("UseZip") ))->SetValue( ! ((wxCheckBox*)(event.GetEventObject()))->IsChecked() );

if ( id == XRCID("UseZip") )
  ((wxCheckBox*)FindWindow( wxT("UseTar") ))->SetValue( ! ((wxCheckBox*)(event.GetEventObject()))->IsChecked() );
}

void MakeArchiveDialog::OnUpdateUI(wxUpdateUIEvent& event)
{
if ( event.GetId() == XRCID("UseTar") )          // If Tar button, do UI for its subordinates
  {  bool enabled = ((wxCheckBox*)event.GetEventObject())->GetValue();
    ((wxCheckBox*)FindWindow( wxT("DeleteSources") ))->Enable(enabled);
    ((wxCheckBox*)FindWindow( wxT("DeReference") ))->Enable(enabled);
    ((wxCheckBox*)FindWindow( wxT("Verify") ))->Enable(enabled);
    ((wxRadioBox*)FindWindow( wxT("CompressRadio") ))->Enable(enabled);
    ((wxStaticText*)FindWindow( wxT("CreateInFolder") ))->Enable(enabled);
    ((wxBitmapButton*)FindWindow( wxT("CreateInBrowse") ))->Enable(enabled);
    if ( createincombo != NULL ) createincombo->Enable(enabled);
  }
    
if ( event.GetId() == XRCID("wxID_OK") )        // Enable the OK button if archive-name present &&  >0 files
  {  if ( NewArchive )   OKEnabled = ( ((wxTextCtrl*)FindWindow( wxT("ArchiveName") ))->GetValue() != wxEmptyString );
      else           OKEnabled = ( createincombo->GetValue() != wxEmptyString );
    ((wxButton*)event.GetEventObject())->Enable( list->GetCount() && OKEnabled );
  }
}

void MakeArchiveDialog::OnOK(wxCommandEvent &event)
{
if ( ! OKEnabled ) return;                                    // If insufficient data to allow wxID_OK, ignore the keypress
                        // If there is a new entry in the create-in comb, add it to the history if it's unique
wxString string = createincombo->GetValue();                  // Get the visible string
if ( ! ( string.IsEmpty() || string==wxT("./") )  ) 
  {  int index = parent->ArchiveHistory.Index( string, false ); // See if this entry already exists in the history array
    if ( index != wxNOT_FOUND )
      parent->ArchiveHistory.RemoveAt( index );               // If so, remove it.  We'll add it back into position zero  
    parent->ArchiveHistory.Insert( string, 0 );               // Either way, insert into position zero
  }

EndModal(wxID_OK);                                            // Let parent take over
}

void MakeArchiveDialog::OnCreateInBrowse(wxCommandEvent& WXUNUSED(event))  // Browse for the dir to store the new archive, or the filepath of existing one
{
wxString filepath;

if ( NewArchive ) 
  {  
    #if wxVERSION_NUMBER >2604 && defined(__WXGTK20__)
      int style = wxDD_DEFAULT_STYLE | wxDD_DIR_MUST_EXIST;     // the gtk2 dirdlg uses different flags :/
    #else
     int style = wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER;
    #endif //wxVERSION_NUMBER >2604 && defined(__WXGTK20__)
    wxDirDialog ddlg(this,_("Choose the Directory in which to store the Archive"), parent->active->GetActiveDirPath(), style );
    if ( ddlg.ShowModal()  != wxID_OK ) return;    
    filepath = ddlg.GetPath();                                // Get the selection
  }  
 else
  {  
    wxFileDialog fdlg(this,_("Browse for the Archive to which to Append"), parent->active->GetActiveDirPath(), wxT(""), wxT("*"), wxFD_OPEN );
    if ( fdlg.ShowModal()  != wxID_OK ) return;    
    filepath = fdlg.GetPath();
  }  

if ( ! filepath.IsEmpty() )   createincombo->SetValue( filepath );
}

enum returntype MakeArchiveDialog::GetCommand()
{
if ( ! list->GetCount() ) return cancel;                      // Shouldn't happen, as Return would be disabled

parent->FilepathArray.Clear();                                // In case files were added/deleted during the dialog, clear the old data
for ( size_t n=0; n < list->GetCount(); ++n )                 //  & replace with up-to-data
  parent->FilepathArray.Insert( list->GetString(n), 0 );

if ( NewArchive )  return GetNewCommand();                    // Are we creating or appending?
 else   return GetAppendCommand();
}

enum returntype MakeArchiveDialog::GetNewCommand()  // Parse entered options into command for a new archive
{
wxString command, archivefilename;

archivefilename = ((wxTextCtrl*)FindWindow( wxT("ArchiveName") ))->GetValue();  // Now get the archive file name

                                  // See which checkboxes are set.
if ( ((wxCheckBox*)FindWindow(wxT("UseZip")))->GetValue() )   // If zip is, that'll be the only thing to look up
  {  wxString oldcwd = wxGetCwd();                            // Save the current cwd
    wxString cwd = parent->active->GetActiveDirPath();
    wxSetWorkingDirectory( cwd );                             // & then set it to the currently-selected path (this solves various zip problems)
    parent->Updatedir.Add(cwd);                               // We'll want to update this dir once we're finished
    
    command = wxT("echo Creating an archive using zip; ");
    wxString changedir; changedir.Printf( wxT("cd \\\"%s\\\" && "), cwd.c_str() );
    command += changedir;
    command += wxT("zip -rTv \\\"");
    
    command += archivefilename; command += wxT("\\\"");       // Add the name of the archive

  // Add the files.  First amputate the cwd path, as unzip assumes we stored paths relative to this, & reconstitutes /home/me/foo as /home/me/home/me/foo
    for ( size_t c=0; c < parent->FilepathArray.GetCount(); ++c )
      {  wxString relativefilepath,  filepath = parent->FilepathArray[c];
        if ( filepath.StartsWith( cwd+wxT('/'), &relativefilepath ) )  // If path found, wString::StartsWith puts the rest of the string in relativefilepath
           filepath = relativefilepath;                       // If successful, this means filepath contains the (file)path relative to cwd
         
        command += wxT(" \\\"") + filepath; command += wxT("\\\"");// Add the filepath, cropped or otherwise, to the command
      }
    
    wxSetWorkingDirectory( oldcwd );                          // Reinstate the original cwd
  }
 else                              // If not zip, it must be tar. Start by finding the archive filepath
   {  wxString archivename = createincombo->GetValue();       // Get first the selected path for the archive
    if ( archivename == wxT("./") )                           // If asking for the 'current' dir, get it from active pane
        archivename = parent->active->GetActiveDirPath();
    if ( archivename.Last() != wxT('/') )  archivename += wxT('/');  // Make the dir into a dir
    if ( !wxDirExists( archivename ) )
      {  wxMessageDialog dialog(this, _("The directory in which you want to create the archive doesn't seem to exist.\nUse the current directory?"), wxT(""),wxYES_NO | wxICON_QUESTION );
        if ( dialog.ShowModal() != wxID_YES )  return retry;
        archivename = parent->active->GetActiveDirPath();     // If user chooses, make archive here
      }
    
    parent->Updatedir.Add(archivename);                       // We'll want to update this dir once we're finished.  Grab the chance to save it
    archivename += archivefilename;                           //   then add it to the path     
    
    archivename += wxT(".tar");                               // Add the ext to archive name
    command = wxT("echo Creating an archive using tar; tar");
    command += MakeTarCommand( archivename );                 // The most of the remainder is done here, to reduce duplication
    
                                  // Now sort out the compression command, if any
    switch( parent->ArchiveCompress = ((wxRadioBox*)FindWindow( wxT("CompressRadio") ))->GetSelection() )
      {  case 0:  command += wxT("&& echo Compressing with bzip2 && bzip2 -z \\\""); command += archivename; command += wxT("\\\"");
                break;
         case 1:  command += wxT("&& echo Compressing with gzip && gzip \\\""); command += archivename; command += wxT("\\\"");
                break;
      }
  }

parent->command = command;                                    // Pass the command string to parent's version
return valid;
}

enum returntype MakeArchiveDialog::GetAppendCommand()  // Parse entered options into command to append to an archive
{
wxString command;
                                  // Start by finding the archive filepath
wxString archivename = createincombo->GetValue();             // Get first the selected path for the archive
if ( !wxFileExists( archivename ) )
  {  wxMessageDialog dialog(this, _("The archive to which you want to append doesn't seem to exist.\nTry again?"), wxT(""),wxYES_NO | wxICON_QUESTION );
    return ( dialog.ShowModal()==wxID_YES ? retry : cancel );
  }

wxString ext = archivename.AfterLast('.'); ext.MakeLower();

if ( ext==wxT("zip") || ext==wxT("htb"))
  {  wxString oldcwd = wxGetCwd();                            // Save the current cwd
    wxString cwd = parent->active->GetActiveDirPath();
    wxSetWorkingDirectory( cwd );                             // & then set it to the currently-selected path (this solves various zip problems)

    command = wxT("echo Appending to a zip archive; ");
    wxString changedir; changedir.Printf( wxT("cd \\\"%s\\\" && "), cwd.c_str() );
    command += changedir;
    command += wxT("zip -urv");                               // Update, recursive, verbose
    
    if ( (parent->DeleteSource = ((wxCheckBox*)FindWindow( wxT("DeleteSources") ))->GetValue()) )  // Do we want to delete the source files?
      command += wxT('m');
    if ( (parent->Verify = ((wxCheckBox*)FindWindow( wxT("Verify") ))->GetValue()) )               // Do we want to verify the archive?
      command += wxT('T');  
    if ( ! (parent->DerefSymlinks = ((wxCheckBox*)FindWindow( wxT("DeReference") ))->GetValue()))  // Do we want to save symlink itself, not the target?
      command += wxT('y');
      
    wxString relativefilepath;          // Zip wants its archive to be relative to the cwd. archivename is probably an absolute filepath
    if ( archivename.StartsWith( cwd+wxT('/'), &relativefilepath ) )// If path found, wString::StartsWith puts the rest of the string in relativefilepath
       archivename = relativefilepath;                              // If successful, this means filepath contains the (file)path relative to cwd
    command += wxT(" \\\"") + archivename; command += wxT("\\\"");  // Add the name of the archive

  // Add the files.  First amputate the cwd path, as unzip assumes we stored paths relative to this, & reconstitutes /home/me/foo as /home/me/home/me/foo
    for ( size_t c=0; c < parent->FilepathArray.GetCount(); ++c )
      {  wxString relativefilepath,  filepath = parent->FilepathArray[c];
        if ( filepath.StartsWith( cwd+wxT('/'), &relativefilepath ) )  // If path found, wString::StartsWith puts the rest of the string in relativefilepath
           filepath = relativefilepath;                                // If successful, this means filepath contains the (file)path relative to cwd

        command += wxT(" \\\"") + filepath; command += wxT("\\\"");    // Add the filepath, cropped or otherwise, to the command
      }
      
    wxSetWorkingDirectory( oldcwd );                          // Reinstate the original cwd
  }

 else if ( ext==wxT("tar") )
   {  command = wxT("echo Appending to a tarball; tar");
    command += MakeTarCommand( archivename );                 // The remainder is done here, to reduce duplication
  }
  
 else if ( ext==wxT("tgz") || ! archivename.Right(7).CmpNoCase( wxT(".tar.gz") ) )  
  {  command = wxT("echo Appending to a tarball compressed with gzip; ");
    command += AppendCompressedTar( archivename, gzip );      // The remainder is done here, to reduce duplication
  }
  
 else if ( ext==wxT("tbz2") || ! archivename.Right(8).CmpNoCase( wxT(".tar.bz2") ) )  
  {  command = wxT("echo Appending to a tarball compressed with bzip2; ");
    command += AppendCompressedTar( archivename, bzip );
  }
 
 else return cancel;    // if archive doesn't seem to be an archive (or maybe it's a .Z, 'Compress'ed one, which I'm ignoring. Compress isn't on my system anyway!)

parent->command = command;                                    // Pass the command string to parent's version
return valid;
}

wxString MakeArchiveDialog::AppendCompressedTar( wxString& archivename, enum ziptype ztype )  // Handles the uncompress, append, recompress contruction
{
wxString command, UncompressedName, Recompress;

UncompressedName = FindUncompressedName( archivename, ztype );  // eg  .tbz2 or .tar.gz -> .tar   
if ( UncompressedName.IsEmpty() ) return UncompressedName;
   
switch( ztype )
  {  case gzip :  command = wxT("gunzip -v \\\"") + archivename; command += wxT("\\\"");
              Recompress =  wxT(" ; echo Recompressing with gzip && gzip \\\"") + UncompressedName; Recompress += wxT("\\\""); break;
    case bzip :  command = wxT("bunzip2 -v \\\"") + archivename; command += wxT("\\\"");
              Recompress =  wxT(" ; echo Recompressing with bzip2 && bzip2 -z \\\"") + UncompressedName; Recompress += wxT("\\\""); break;
     default: return command;
  }

command += wxT(" && tar");
command += MakeTarCommand( UncompressedName );
command += Recompress;

return command;
}

wxString MakeArchiveDialog::MakeTarCommand( wxString& archivename )
{
wxString command;

if ( (parent->DeleteSource = ((wxCheckBox*)FindWindow( wxT("DeleteSources") ))->GetValue()) )  // Do we want to delete the source files?
  command += wxT(" --remove-files");

if ( NewArchive )  command += wxT(" -cv");                    // -c designates new archive, -v is verbose
 else  command += wxT(" -rv");                                // -r is append

if ( (parent->DerefSymlinks = ((wxCheckBox*)FindWindow( wxT("DeReference") ))->GetValue()) )  // Do we want to save symlink target, not the link itself?
  command += wxT('h');

command += wxT("f \\\""); command += archivename + wxT("\\\"");// Add the name of the archive

command += AddSortedFiles();                                   // Add the files, sorted by dir, each dir preceded by a -C

if ( (parent->Verify = ((wxCheckBox*)FindWindow( wxT("Verify") ))->GetValue()) )  // Verify the archive? NB we do this here, not using the 'W' flag.
  {                                                                               // That's because it would fail if the source files had just been deleted
    wxString str; str.Printf( wxT("&& echo Verifying %s && "), archivename.AfterLast(wxT('/')).c_str() );  
    command += str;
    command += wxT("tar -tvf \\\""); command += archivename + wxT("\\\"");
  }

return command;
}

wxString MakeArchiveDialog::AddSortedFiles()  // Returns files sorted by dir, each dir preceded by a -C
{
wxString command, lastpath;
wxArrayString filepaths, files;

if ( ! parent->FilepathArray.GetCount() ) return command;

parent->FilepathArray.Sort();                                         // This should get them into dir order

for ( size_t c=0; c < parent->FilepathArray.GetCount(); ++c )         // It's easier if we separate the 'local' files from the filepaths
  if ( parent->FilepathArray[c].Left(2) == wxT("./") )                // Initial './', so this should be a file in the cwd. Remove its ./
            files.Add( parent->FilepathArray[c].Mid(2) );
   else if ( parent->FilepathArray[c].GetChar(0) != wxT('/') ) files.Add( parent->FilepathArray[c] );  // No initial '/', so this should be a file in the cwd
    else filepaths.Add( parent->FilepathArray[c] );                   // Otherwise put it in the filepaths array 
  
if ( files.GetCount() )          // Because the cwd probably isn't the active dir, we need to use the latter explicitly
  { command = wxT(" -C \\\""); command += parent->active->GetActiveDirPath(); command += wxT("\\\""); }
for ( size_t c=0; c < files.GetCount(); ++c )
  { command += wxT(" \\\"") + files[c];  command += wxT("\\\""); }    // Add each 'local' file to the command

size_t n=0;
if ( filepaths.GetCount() )
  do                                                                  // Now do the absolute filepaths, path by path
  {  lastpath = filepaths[n].BeforeLast('/');                         // Start with the path, making a new -C entry
    command += wxT(" -C \\\""); command += lastpath; command += wxT("/\\\"");
    do
      {  command += wxT(" \\\"") + filepaths[n++].AfterLast('/');     // Add each file for this path
        command += wxT("\\\"");
        if ( n >= filepaths.GetCount() ) return command;              // Return if we're finished
      }
      while ( filepaths[n].BeforeLast('/') == lastpath );             // Continue while the path is unchanged
  }  
   while (1);                                                         // The path changed, so loop to cd

return command;
}

BEGIN_EVENT_TABLE(MakeArchiveDialog,ArchiveDialogBase)
  EVT_CHECKBOX(-1, MakeArchiveDialog::OnCheckBox)
  EVT_BUTTON(XRCID("CreateInBrowse"), MakeArchiveDialog::OnCreateInBrowse)
  EVT_BUTTON(wxID_OK, MakeArchiveDialog::OnOK)
  EVT_UPDATE_UI(-1, MakeArchiveDialog::OnUpdateUI)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------------------------------------------------

void CompressDialog::Init( Archive* dad )
{
init( dad );

size_t count = parent->FilepathArray.GetCount();
for ( size_t n=0; n < count; ++n )                                    // Put the highlit files into the listbox
  if ( ! (parent->FilepathArray[n]==parent->active->GetActiveDirectory() && count==1) )  // We don't want rootdir in the list just because it's automatically highlit
    list->Append( parent->FilepathArray[n] );                         // Put in the files-to-be-compressed listbox

                                // Set checkboxes etc to last-used values
((wxSlider*)FindWindow( wxT("Slider") ))->SetValue( parent->Slidervalue );
((wxRadioBox*)FindWindow( wxT("CompressRadio") ))->SetSelection( parent->Compressionchoice );
((wxCheckBox*)FindWindow( wxT("Recurse") ))->SetValue( parent->Recurse );
((wxCheckBox*)FindWindow( wxT("Force") ))->SetValue( parent->Force );
}

void CompressDialog::OnUpdateUI(wxUpdateUIEvent& event)
{
if ( event.GetId() == XRCID("Slider") )                               // I'm using the slider UI event, as the radiobox didn't seem to generate one!?  
  {  bool enabled = ((wxRadioBox*)FindWindow( wxT("CompressRadio") ))->GetSelection();  // If which-compressor radio isn't gzip (==1), disable the slider
    ((wxStaticText*)FindWindow( wxT("FasterStatic") ))->Enable(enabled);
    ((wxSlider*)FindWindow( wxT("Slider") ))->Enable(enabled);
    ((wxStaticText*)FindWindow( wxT("SmallerStatic") ))->Enable(enabled);
  }
    
if ( event.GetId() == XRCID("wxID_OK"))                               // Enable the OK button if >0 files in listbox
    ((wxButton*)event.GetEventObject())->Enable( list->GetCount() );
}

enum returntype CompressDialog::GetCommand()    // Parse entered options into a command
{
wxString command, files, verifystring;
bool data = 0;                                             // Flags that we've got >0 files to compress

if ( ! list->GetCount() ) return cancel;                   // Shouldn't happen, as Return would be disabled

parent->FilepathArray.Clear();                             // In case files were added/deleted during the dialog, clear the old data
for ( size_t n=0; n < list->GetCount(); ++n )              //  & replace with up-to-data
    parent->FilepathArray.Insert( list->GetString(n), 0 );

bool UseGzip = ((wxRadioBox*)FindWindow( wxT("CompressRadio") ) )->GetSelection();  // Which compressor are we using?
parent->Compressionchoice = UseGzip;

command = wxT("echo Compressing ");
switch( UseGzip )                                          // Put the announcement & program invocation into command
  {  case 0:  command += wxT("using bzip2; bzip2 -zv"); break; // Compress, verbose
    case 1:  command += wxT("using gzip; gzip -v");
            wxString squash; squash.Printf(wxT("%u"),  parent->Slidervalue = ((wxSlider*)FindWindow( wxT("Slider") ))->GetValue() );
            command += squash; break;                      // Add the degree-of-compression value, 1-9.  No point doing this for bzip2, which always does 9
  }

if ( (parent->Force = ((wxCheckBox*)FindWindow( wxT("Force") ))->IsChecked()) )
  command += wxT('f');                                     // Force overwrite if requested

bool recurse = ((wxCheckBox*)FindWindow( wxT("Recurse") ))->IsChecked();
parent->Recurse = recurse;

for ( size_t n=0; n < parent->FilepathArray.GetCount(); ++n )       // Now add the requested files
  {  FileData fd( parent->FilepathArray[n] );
    if ( fd.IsDir() )                                               // If this 'file' is a directory, we have to be clever
      {  if ( recurse )   // If we're supposed to, recurse.  NB We had to do this for bz2 as bzip2 has no recurse flag, but my gzip's recursion is buggy anyway
          {  wxArrayString filearray;                          
            AddRecursively( parent->FilepathArray[n], filearray );  // Recursively put files from this dir & all its subdirs into filearray            
            for ( size_t c=0; c < filearray.GetCount(); ++c )       // Then add them 1 by 1 to the files string
              {  files += wxT(" \\\""); files += filearray[c]; files += wxT("\\\""); data = 1; }
          }    
      }
                                      
    else        // If the file IS a file, or if a dir using gzip -r, add it to the string.  NB we quote using escaped "", in case a filepath contains a '
       { files += wxT(" \\\""); files += parent->FilepathArray[n]; files += wxT("\\\""); data = 1; }
  }

if ( ! data ) return cancel;                               // Lest there were no actual files to compress

parent->command = command + files;                         // Join the command strings in parent's version
return valid;
}

void CompressDialog::AddRecursively( wxString& dir, wxArrayString& filearray )  // Recursively put files from this dir & all its subdirs into filearray
{
wxArrayString localarray;    
                              // First get all the files/dirs from the passed dir into a local array
wxDir::GetAllFiles( dir, &localarray, wxEmptyString, wxDIR_DIRS | wxDIR_FILES | wxDIR_HIDDEN );
for ( size_t c=0; c < localarray.GetCount(); ++c )         // Now go thru them 1 by 1
  {  FileData fd( localarray[c] );
    if ( fd.IsDir() )                   AddRecursively(localarray[c], localarray);  // recursing into dirs (which will append their contents)
       else  filearray.Add( localarray[c] );                                        //  & adding files to the passed array
  }
}

BEGIN_EVENT_TABLE(CompressDialog,ArchiveDialogBase)
  EVT_UPDATE_UI(-1, CompressDialog::OnUpdateUI)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------------------------------------------------

void DecompressDialog::Init( Archive* dad )
{
init( dad );                                               // Do the base-class init first

size_t count = parent->FilepathArray.GetCount();
for ( size_t n=0; n < count; ++n )                         // Put the highlit files into the listbox
  {  enum ziptype ztype = parent->Categorise( parent->FilepathArray[n] );
    if ( ztype==gzip || ztype==bzip                        // If this is a compressed file,
                        || ztype==targz || ztype==tarbz )  //   or a compressed archive,
      list->Append( parent->FilepathArray[n] );            //     put it in the to-be-decompressed listbox
  }
                                // Set checkboxes etc to last-used values
((wxCheckBox*)FindWindow( wxT("Recurse") ))->SetValue( parent->DecompressRecurse );
((wxCheckBox*)FindWindow( wxT("Force") ))->SetValue( parent->DecompressForce );
((wxCheckBox*)FindWindow( wxT("ArchivesToo") ))->SetValue( parent->DecompressArchivesToo );
}

void DecompressDialog::OnUpdateUI(wxUpdateUIEvent& event)
{
if ( event.GetId() == XRCID("wxID_OK") )                   // Enable the OK button if >0 files in listbox
    ((wxButton*)event.GetEventObject())->Enable( list->GetCount() );
}

enum returntype DecompressDialog::GetCommand()  // Parse entered options into a command
{
wxString command; wxArrayString gz, bz;

if ( ! list->GetCount() ) return cancel;                   // Shouldn't happen, as Return would be disabled

parent->FilepathArray.Clear();                             // In case files were added/deleted during the dialog, clear the old data
for ( size_t n=0; n < list->GetCount(); ++n )              //  & replace with up-to-data
  parent->FilepathArray.Insert( list->GetString(n), 0 );

bool recurse = ((wxCheckBox*)FindWindow( wxT("Recurse") ))->IsChecked();
parent->DecompressRecurse = recurse;
bool force = ((wxCheckBox*)FindWindow( wxT("Force") ))->IsChecked();
parent->DecompressForce = force;
parent->DecompressArchivesToo = ((wxCheckBox*)FindWindow( wxT("ArchivesToo") ))->IsChecked();


    // We may have 3 sorts of file in the array:  gz, bz2, and ordinary files added in error.  Oh, and dirs. Because of possible recursion, do in a submethod
if ( ! SortFiles( parent->FilepathArray, gz, bz, recurse, parent->DecompressArchivesToo ) )        
  {  wxMessageDialog dialog(this, _("No relevant compressed files were selected.\nTry again?"),    // If there weren't any compressed files selected . . .
                                                                        wxT(""),wxYES_NO | wxICON_QUESTION );
    if ( dialog.ShowModal() == wxID_YES )  return retry; else return cancel;
  }
  
command = wxT("echo Decompressing; ");
if ( ! gz.IsEmpty() )                                      // Create a gunzip command, adding relevant files
  {  command += wxT("gunzip -v");                          // Verbose
    if ( force ) command += wxT('f');
    for ( size_t c=0; c < gz.GetCount(); ++c )             // Add files to the command
      {  command += wxT(" \\\""); command += gz[c]; command += wxT("\\\""); }
    command += wxT(';');
  }

if ( ! bz.IsEmpty() )
  {  command += wxT("bunzip2 -v"); 
    if ( force ) command += wxT('f');
    for ( size_t c=0; c < bz.GetCount(); ++c )
      {  command += wxT(" \\\""); command += bz[c]; command += wxT("\\\""); }
  }

parent->command = command;  
return valid;
}

bool DecompressDialog::SortFiles( wxArrayString& selected, wxArrayString& gz, wxArrayString& bz, bool recurse, bool archivestoo /*=false */)  //  Sort selected files into .bz, .gz & rubbish
{
for ( size_t n=0; n < selected.GetCount(); ++n )           // Go thru the selected files
  {  FileData fd( selected[n] );
    if ( fd.IsDir() )                                      // If this 'file' is a directory, we have to be clever
      {  if ( ! recurse ) continue;                        // If we're not supposed to recurse, ignore this item

        wxArrayString filearray;      // Recursively put files from this dir & all its subdirs into filearray
        wxDir::GetAllFiles( selected[n], &filearray, wxEmptyString, wxDIR_DIRS | wxDIR_FILES | wxDIR_HIDDEN );
        SortFiles( filearray, gz, bz, true, archivestoo ); // Now sort by recursion
        continue;                                          // We've processed this item. Loop here to avoid falling thru to the next bit
      }    
    
    switch( parent->Categorise( selected[n] ) )            // This method returns different answers depending on whether .gz, .tar.bz2 etc
      {  case targz:  if ( !archivestoo )  break;          // If we don't want to decompress any compressed archives, break. Otherwise fall into gzip
        case gzip:  gz.Add( selected[n] ); break;          // Add the item to the appropriate string
        case tarbz:  if ( !archivestoo )  break;
        case bzip:  bz.Add( selected[n] ); break;
          default: break;                                  // because the file is not compressed
      }
  }
return ! ( gz.IsEmpty() && bz.IsEmpty() );
}



BEGIN_EVENT_TABLE(DecompressDialog,ArchiveDialogBase)
  EVT_UPDATE_UI(-1, DecompressDialog::OnUpdateUI)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------------------------------------------------

void ExtractArchiveDlg::Init( Archive* dad )
{
parent = dad;

text = (wxTextCtrl*)FindWindow( wxT("text") );
CreateInCombo = (wxComboBox*)FindWindow( wxT("CreateInCombo") );

for ( size_t n=0; n < parent->ArchiveHistory.GetCount(); ++n )    // Load the archive path history into the combo
  CreateInCombo->Append( parent->ArchiveHistory[n] );
CreateInCombo->SetValue(wxT("./"));                               // Give archive-path a default value of ./

parent->active->GetMultiplePaths( parent->FilepathArray );        // Load the selected files into the stringarray

size_t count = parent->FilepathArray.GetCount();
for ( size_t n=0; n < count; ++n )
  if ( ArchiveDialogBase::IsAnArchive( parent->FilepathArray[n] ) )
      { text->ChangeValue( parent->FilepathArray[n] ); break; }   // Put 1st archive we find into the textctrl, then stop looking

                                // Set checkbox to last-used value
((wxCheckBox*)FindWindow( wxT("Force") ))->SetValue( parent->ArchiveExtractForce );
}

void ExtractArchiveDlg::OnFileBrowse(wxCommandEvent& WXUNUSED(event))  // Browse for archive
{
wxFileDialog fdlg(this,_("Browse for the Archive to Verify"), parent->active->GetActiveDirPath(), wxT(""), wxT("*"), wxFD_OPEN );
if ( fdlg.ShowModal()  != wxID_OK ) return;    
  
wxString filepath = fdlg.GetPath();
if ( ! filepath.IsEmpty() )
  text->ChangeValue( filepath );
}

void ExtractArchiveDlg::OnCreateInBrowse(wxCommandEvent& WXUNUSED(event))  // Browse for the dir into which to extract the archive
{
#if wxVERSION_NUMBER > 2604 && defined(__WXGTK20__)
  int style = wxDD_DEFAULT_STYLE | wxDD_DIR_MUST_EXIST;      // the gtk2 dirdlg uses different flags :/
#else
  int style = wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER;
#endif //wxVERSION_NUMBER >2604 && defined(__WXGTK20__)

wxDirDialog ddlg(this,_("Choose the Directory in which to Extract the archive"), parent->active->GetActiveDirPath(), style );
if ( ddlg.ShowModal()  != wxID_OK ) return;    

wxString filepath = ddlg.GetPath();                               // Get the selection
if ( ! filepath.IsEmpty() )   CreateInCombo->SetValue( filepath );
}

void ExtractArchiveDlg::OnUpdateUI(wxUpdateUIEvent& event)
{
if ( event.GetId() == XRCID("wxID_OK") )                          // Enable the OK button if archive name in textctrl
    ((wxButton*)event.GetEventObject())->Enable( ! text->GetValue().IsEmpty() );
}

void ExtractArchiveDlg::OnOK(wxCommandEvent &event)
{
if ( ! ((wxButton*)event.GetEventObject())->IsEnabled() ) return;// If insufficient data to allow wxID_OK, ignore the keypress
                        // If there is a new entry in the create-in comb, add it to the history if it's unique
wxString string = CreateInCombo->GetValue();                     // Get the visible string
if ( ! ( string.IsEmpty() || string==wxT("./") )  ) 
  {  int index = parent->ArchiveHistory.Index( string, false );  // See if this entry already exists in the history array
    if ( index != wxNOT_FOUND )
      parent->ArchiveHistory.RemoveAt( index );                  // If so, remove it.  We'll add it back into position zero  
    parent->ArchiveHistory.Insert( string, 0 );                  // Either way, insert into position zero
  }
  
EndModal(wxID_OK);
}


enum returntype ExtractArchiveDlg::GetCommand()
{
wxString command, decompress;

wxString createindir = CreateInCombo->GetValue();                // Get first the selected path for the archive
if ( createindir == wxT("./") )  createindir = parent->active->GetActiveDirPath();

if ( !wxDirExists( createindir ) )
  if ( !wxFileName::Mkdir( createindir, 0777, wxPATH_MKDIR_FULL ))// If it doesn't exist, create it.  The flag means that ./newsubdir/anothernewdir/yetanother works
      {  wxMessageDialog dialog(this, _("Failed to create the desired destination directory.\nTry again?"), wxT(""),wxYES_NO | wxICON_QUESTION );
        if ( dialog.ShowModal() == wxID_YES )  return retry; else return cancel;
      }

parent->Updatedir.Add(createindir);                              // We'll want to update this dir once we're finished.  Grab the chance to save it

bool force = ((wxCheckBox*)FindWindow( wxT("Force") ))->GetValue();
parent->ArchiveExtractForce = force;

wxString archive = text->GetValue();                             // Get the archive filepath

enum ziptype ztype = parent->Categorise( archive );              // Is it a bird, is it a plane . . .?
switch( ztype )
  {   case htb:                                                  // htb's are effectively zips  
      case zip:      command.Printf( wxT("echo Unzipping %s;"), archive.AfterLast(wxT('/')).c_str() );
                  command += wxT("unzip ");
                  command += ( force ? wxT("-o") : wxT("-n") );
                  command += wxT(" \\\""); command += archive; command += wxT("\\\""); // Add the archive filepath
                  if ( !createindir.IsEmpty() )                  // Extract to the destination path, if one was requested
                    {  wxString destdir;   destdir.Printf( wxT(" -d \\\"%s\\\" "), createindir.c_str() );
                      command += destdir;
                    }  
                  break;

      case targz:    decompress = wxT('z');
      case tarbz:    if ( ztype==tarbz )  decompress = wxT('j'); // The 'if' is necessary because we may be coming from targz above
      case taronly:  command.Printf( wxT("echo Extracting %s;"), archive.AfterLast(wxT('/')).c_str() );
                  if ( !createindir.IsEmpty() )                  // cd to the destination path, if one was requested
                    {  wxString changedir;   changedir.Printf( wxT("cd \\\"%s\\\" && "), createindir.c_str() );
                      command += changedir;
                    }  
                  command += wxT("tar ");
                  if ( force )  command += wxT("--overwrite ");
                  command += wxT(" -xv");
                  if ( ! force )  command += wxT('k');
                  command += decompress;                         // Decompress may be z, j, or ""
                  command += wxT("f \\\"");
                  command += archive + wxT("\\\"");              // Add the archive filepath
                  break;
       
      default:    wxMessageDialog dialog(this, _("Can't find a valid archive to extract.\nTry again?"), wxT(""),wxYES_NO | wxICON_QUESTION );
                  if ( dialog.ShowModal() == wxID_YES )  return retry; else return cancel;
  }

parent->command = command;                                      // Pass the command string to parent's version
return valid;
}


BEGIN_EVENT_TABLE(ExtractArchiveDlg,wxDialog)
  EVT_BUTTON(XRCID("Browse"), ExtractArchiveDlg::OnFileBrowse)
  EVT_BUTTON(XRCID("CreateInBrowse"), ExtractArchiveDlg::OnCreateInBrowse)
  EVT_BUTTON(wxID_OK, ExtractArchiveDlg::OnOK)
  EVT_UPDATE_UI(-1, ExtractArchiveDlg::OnUpdateUI)
END_EVENT_TABLE()
//-----------------------------------------------------------------------------------------------------------------------

void VerifyCompressedDialog::Init( Archive* dad )
{
init( dad );

size_t count = parent->FilepathArray.GetCount();
for ( size_t n=0; n < count; ++n )                               // Put the highlit files into the listbox
  {  enum ziptype ztype = parent->Categorise( parent->FilepathArray[n] );
    if ( ztype==gzip || ztype==bzip )                            // If this is a compressed file,
      list->Append( parent->FilepathArray[n] );                  //   put it in the files-to-be-verified listbox
  }
}

enum returntype VerifyCompressedDialog::GetCommand()  // Parse entered options into a command
{
wxString command; wxArrayString gz, bz;

if ( ! list->GetCount() ) return cancel;                         // Shouldn't happen, as Return would be disabled

parent->FilepathArray.Clear();                                   // In case files were added/deleted during the dialog, clear the old data
for ( size_t n=0; n < list->GetCount(); ++n )                    //  & replace with up-to-data
  parent->FilepathArray.Insert( list->GetString(n), 0 );

        // We may have 3 sorts of file in the array:  gz, bz2, and ordinary files added in error
if ( ! SortFiles( parent->FilepathArray, gz, bz, false ) )       // If there weren't any compressed files selected . . .
  {  wxMessageDialog dialog(this, _("No relevant compressed files were selected.\nTry again?"), wxT(""),wxYES_NO | wxICON_QUESTION );
    if ( dialog.ShowModal() == wxID_YES )  return retry; else return cancel;
  }
  
command = wxT("echo Verifying; ");
if ( ! gz.IsEmpty() )                                            // Create a gunzip command, adding relevant files
  {  command += wxT("gunzip -tv");                               // Test, Verbose
    for ( size_t c=0; c < gz.GetCount(); ++c )                   // Add files to the command
      {  command += wxT(" \\\""); command += gz[c]; command += wxT("\\\""); }
    command += wxT(';');
  }

if ( ! bz.IsEmpty() )
  {  command += wxT("bunzip2 -tv"); 
    for ( size_t c=0; c < bz.GetCount(); ++c )
      {  command += wxT(" \\\""); command += bz[c]; command += wxT("\\\""); }
  }

parent->command = command;  
return valid;
}

//-----------------------------------------------------------------------------------------------------------------------
void VerifyArchiveDlg::Init( Archive* dad )
{
parent = dad;

text = (wxTextCtrl*)FindWindow( wxT("text") );

parent->active->GetMultiplePaths( parent->FilepathArray );       // Load the selected files into the stringarray

size_t count = parent->FilepathArray.GetCount();
for ( size_t n=0; n < count; ++n )
  if ( ArchiveDialogBase::IsAnArchive( parent->FilepathArray[n] ) )
      { text->AppendText( parent->FilepathArray[n] ); break; }   // Put 1st archive we find into the textctrl, then stop looking
}

void VerifyArchiveDlg::OnOK(wxCommandEvent &event)
{
if ( ! ((wxButton*)event.GetEventObject())->IsEnabled() ) return;  // If insufficient data to allow wxID_OK, ignore the keypress

EndModal(wxID_OK);
}

enum returntype VerifyArchiveDlg::GetCommand()
{
wxString command, archive, Recompress, str;

wxString archivename = text->GetValue();                   // Get the archive filepath

enum ziptype ztype = parent->Categorise( archivename );    // Is it a bird, is it a plane . . .?

wxString UncompressedName = ArchiveDialogBase::FindUncompressedName( archivename, ztype );  // eg  .tbz2 or .tar.gz -> .tar 
if ( ! UncompressedName.IsEmpty() )                        // If there's a name, this must be a compressed tar
      archive = UncompressedName;                          // So we make sure the name that Tar is fed is the uncompressed one
 else  archive = archivename;

switch( ztype )
  {  case htb:
    case zip:        command.Printf( wxT("echo Verifying %s;"), archivename.AfterLast(wxT('/')).c_str() );
                  command += wxT("unzip -t");
                  command += wxT(" \\\"");  command += archivename + wxT("\\\"");  // Add the archive filepath
                  break;

    case targz:      command = wxT("gunzip \\\"") + archivename; command += wxT("\\\" && ");
                  Recompress =  wxT("; gzip \\\"") + UncompressedName; Recompress += wxT("\\\"");
     case tarbz:      if ( ztype==tarbz )                  // The 'if' is necessary because we may be coming from targz above
                    { command = wxT("bunzip2 \\\"") + archivename;  command += wxT("\\\" && ");
                      Recompress =  wxT("; bzip2 -z \\\"") + UncompressedName; Recompress += wxT("\\\"");
                    }
     case taronly:    str.Printf( wxT("echo Verifying %s && "), archivename.AfterLast(wxT('/')).c_str() );  // archivename means if compressed, the original name is used
                  command += str;
                  command += wxT("tar ");
                  command += wxT("-tv");
                  command += wxT("f \\\"");
                  command += archive + wxT("\\\"");        // Add the archive filepath.  If originally compressed, uses the uncompressed name
                  command += Recompress;                   // If originally compressed, this will recompress it
                  break;
       
        default:    wxMessageDialog dialog(this, _("Can't find a valid archive to verify.\nTry again?"), wxT(""),wxYES_NO | wxICON_QUESTION );
                  if ( dialog.ShowModal() == wxID_YES )  return retry; else return cancel;
  }

parent->command = command;                                 // Pass the command string to parent's version
return valid;
}

BEGIN_EVENT_TABLE(VerifyArchiveDlg,ExtractArchiveDlg)
  EVT_BUTTON(wxID_OK, VerifyArchiveDlg::OnOK)
  EVT_UPDATE_UI(-1, ExtractArchiveDlg::OnUpdateUI)
END_EVENT_TABLE()
//-----------------------------------------------------------------------------------------------------------------------

void Archive::CreateOrAppend( bool NewArchive /*= true*/ )
{
enum returntype ans;

LoadHistory();

MakeArchiveDialog dlg;
if ( NewArchive )
      wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("MakeArchiveDlg"));
 else  wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("AppendArchiveDlg"));
dlg.Init( this, NewArchive );

do
  {  command.Clear();                                      // Make sure there's no data left over from a previous loop
    if ( dlg.ShowModal() != wxID_OK )
      { ans = cancel; break; }
    
    ans = dlg.GetCommand();                                // Get the dialog to parse the options into a command
    if ( ans == retry )  continue;                         // Give the user the opportunity to correct errors if desired
    break;
  }
 while (1);

if ( ans == cancel )  { delete this; return; }

SaveHistory();

command = wxT("sh -c \"") + command; command += wxT('\"'); // Since command is multiple (if only with echos) we need to execute it thru a shell

            // We now have the command.  Do the dialog that runs it and displays the results
wxBusyCursor busy;

CommandDisplay OutputDlg;
wxXmlResource::Get()->LoadDialog(&OutputDlg, MyFrame::mainframe, wxT("OutputDlg"));
OutputDlg.Init( command );
OutputDlg.ShowModal();

if ( !OutputDlg.exitstatus ) 
  {  if ( NewArchive )  BriefLogStatus bls( _("Archive created"));
      else          BriefLogStatus bls( _("File(s) added to Archive"));
  }
 else  
  {  if ( NewArchive )  BriefLogStatus bls( _("Archive creation failed"));
      else          BriefLogStatus bls( _("Archive addition failed"));
  }

UpdatePanes();
delete this;
}

void Archive::Compress()    // Does *zip on the selected files
{
enum returntype ans;

CompressDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("CompressDlg"));
dlg.Init( this );

do
  {  command.Clear();                                      // Make sure there's no data left over from a previous loop
    if ( dlg.ShowModal() != wxID_OK )
      { ans = cancel; break; }
    
    ans = dlg.GetCommand();                                // Get the dialog to parse the options into a command
    if ( ans == retry )  continue;                         // Give the user the opportunity to correct errors if desired
    break;
  }
 while (1);

if ( ans == cancel )  { delete this; return; }

command = wxT("sh -c \"") + command; command += wxT('\"'); // Since command is multiple (if only with echos) we need to execute it thru a shell

            // We now have the command.  Do the dialog that runs it and displays the results
wxBusyCursor busy;

CommandDisplay OutputDlg;
wxXmlResource::Get()->LoadDialog(&OutputDlg, MyFrame::mainframe, wxT("OutputDlg"));
OutputDlg.Init( command );
OutputDlg.ShowModal();

if ( !OutputDlg.exitstatus ) BriefLogStatus bls( _("File(s) compressed"));
 else    BriefLogStatus bls( _("Compression failed"));

UpdatePanes();
delete this;
}

void Archive::ExtractOrVerify( bool extract /*=true*/ )  // Extracts or verifies either archives or compressed files, depending on which are selected
{
    // Are we decompressing or extracting?  If there is even 1 compressed non-archive file selected, Decompress
    // If there are only archives (+/- rubbish), then select the 1st archive (there should only be one anyway) to extract

active->GetMultiplePaths( FilepathArray );                 // Load the selected files into the stringarray

bool archivefound = false;
if ( ! FilepathArray.IsEmpty() )
  {  for ( size_t n=0; n < FilepathArray.GetCount(); ++n )
      {  enum ziptype ztype = Categorise( FilepathArray[n] );
        if ( ztype==gzip || ztype==bzip )                  // We've found a compressed non-archive so Decompress/Verify this/these
          {  if ( extract )  return Decompress(); 
              else return VerifyCompressed();
          }
         else if ( ztype != NotArchive ) archivefound = true;
      }
    if ( archivefound )                                    // We found no compressed non-archives, & >0 archives, so extract/verify this
      {  if ( extract )  return ExtractArchive(); 
          else return VerifyArchive();
      }
  }
  
          // If we're here, either there were no files selected, or they were all rubbish.  Ask for directions
DecompressOrExtractDlg dlg;
if ( extract )
  wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("DecompressOrExtractDlg"));
 else
  wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("VerifyWhichDlg"));  // VerifyWhichDlg is derived from the above; only labels are different
int ans = dlg.ShowModal();
 
if ( ans==XRCID("Archive") )
  {  if ( extract )  return ExtractArchive(); 
      else return VerifyArchive();
  }

if ( ans==XRCID("Compressed") )
  {  if ( extract )  return Decompress(); 
      else return VerifyCompressed();
  }

delete this;                                               // If neither choice was taken, abort
}

void Archive::Decompress()  // Decompresses files
{
enum returntype ans;

DecompressDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("ExtractCompressedDlg"));
dlg.Init( this );

do
  {  command.Clear();                                      // Make sure there's no data left over from a previous loop
    if ( dlg.ShowModal() != wxID_OK )
      { ans = cancel; break; }
    
    ans = dlg.GetCommand();                                // Get the dialog to parse the options into a command
    if ( ans == retry )  continue;                         // Give the user the opportunity to correct errors if desired
    break;
  }
 while (1);

if ( ans == cancel )   return;

command = wxT("sh -c \"") + command; command += wxT('\"'); // Since command is multiple (if only with echos) we need to execute it thru a shell

            // We now have the command.  Do the dialog that runs it and displays the results
wxBusyCursor busy;
            
CommandDisplay OutputDlg;
wxXmlResource::Get()->LoadDialog(&OutputDlg, MyFrame::mainframe, wxT("OutputDlg"));
OutputDlg.Init( command );
OutputDlg.ShowModal();

if ( !OutputDlg.exitstatus ) BriefLogStatus bls( _("File(s) decompressed"));
 else    BriefLogStatus bls( _("Decompression failed"));

UpdatePanes();
delete this;
}

void Archive::ExtractArchive()  // Extract an Archive
{
enum returntype ans;

LoadHistory();

ExtractArchiveDlg dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("ExtractArchiveDlg"));
dlg.Init( this );

do
  {  command.Clear();                                      // Make sure there's no data left over from a previous loop
    if ( dlg.ShowModal() != wxID_OK )
      { ans = cancel; break; }
    
    ans = dlg.GetCommand();                                // Get the dialog to parse the options into a command
    if ( ans == retry )  continue;                         // Give the user the opportunity to correct errors if desired
    break;
  }
 while (1);

if ( ans == cancel )   return;

SaveHistory();

command = wxT("sh -c \"") + command; command += wxT('\"'); // Since command is multiple (if only with echos) we need to execute it thru a shell

            // We now have the command.  Do the dialog that runs it and displays the results
wxBusyCursor busy;
            
CommandDisplay OutputDlg;
wxXmlResource::Get()->LoadDialog(&OutputDlg, MyFrame::mainframe, wxT("OutputDlg"));
OutputDlg.Init( command );
OutputDlg.ShowModal();

if ( !OutputDlg.exitstatus ) BriefLogStatus bls( _("Archive extracted"));
 else    BriefLogStatus bls( _("Extraction failed"));

UpdatePanes();
delete this;
}

void Archive::VerifyCompressed()
{
enum returntype ans;

VerifyCompressedDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("VerifyCompressedDlg"));
dlg.Init( this );

do
  {  command.Clear();                                      // Make sure there's no data left over from a previous loop
    if ( dlg.ShowModal() != wxID_OK )
      { ans = cancel; break; }
    
    ans = dlg.GetCommand();                                // Get the dialog to parse the options into a command
    if ( ans == retry )  continue;                         // Give the user the opportunity to correct errors if desired
    break;
  }
 while (1);

if ( ans == cancel )   return;

command = wxT("sh -c \"") + command; command += wxT('\"'); // Since command is multiple (if only with echos) we need to execute it thru a shell

            // We now have the command.  Do the dialog that runs it and displays the results
wxBusyCursor busy;

CommandDisplay OutputDlg;
wxXmlResource::Get()->LoadDialog(&OutputDlg, MyFrame::mainframe, wxT("OutputDlg"));
OutputDlg.Init( command );
OutputDlg.ShowModal();

if ( !OutputDlg.exitstatus ) BriefLogStatus bls( _("File(s) Verified"));
 else    BriefLogStatus bls( _("Verification failed"));

delete this;
}

void Archive::VerifyArchive()
{
enum returntype ans;

VerifyArchiveDlg dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("VerifyArchiveDlg"));
dlg.Init( this );

do
  {  command.Clear();                                      // Make sure there's no data left over from a previous loop
    if ( dlg.ShowModal() != wxID_OK )
      { ans = cancel; break; }
    
    ans = dlg.GetCommand();                                // Get the dialog to parse the options into a command
    if ( ans == retry )  continue;                         // Give the user the opportunity to correct errors if desired
    break;
  }
 while (1);

if ( ans == cancel )   return;

command = wxT("sh -c \"") + command; command += wxT('\"'); // Since command is multiple (if only with echos) we need to execute it thru a shell

            // We now have the command.  Do the dialog that runs it and displays the results
wxBusyCursor busy;

CommandDisplay OutputDlg;
wxXmlResource::Get()->LoadDialog(&OutputDlg, MyFrame::mainframe, wxT("OutputDlg"));
OutputDlg.Init( command );
OutputDlg.ShowModal();

if ( !OutputDlg.exitstatus ) BriefLogStatus bls( _("Archive Verified"));
 else    BriefLogStatus bls( _("Verification failed"));

delete this;
}

//static
enum ziptype Archive::Categorise( wxString& ftype )    // Returns the type of file in filetype ie .gz, .tar etc
{
if ( ftype.IsEmpty() || ftype.Last() == wxFILE_SEP_PATH ) return NotArchive;

wxString endExt, prevExt, filetype = ftype.AfterLast(wxFILE_SEP_PATH);  // We're only interested in the current filename, not a dir with a '.' halfway up the path
                                  
endExt = filetype.AfterLast( wxT('.') );
if ( endExt == filetype ) return NotArchive;               // Check there IS an ext (AfterLast() returns whole string if the char isn't found)

prevExt = (filetype.BeforeLast( wxT('.') )).AfterLast( wxT('.') ); // Get any previous ext, in case of foo.tar.gz

if ( endExt==wxT("htb") )  return htb;
if ( endExt==wxT("zip") )    return zip;

if ( endExt==wxT("tar") )    return taronly;

if ( endExt==wxT("tgz") )  return targz;
if ( prevExt==wxT("tar") && endExt==wxT("gz") )  return targz;

if ( endExt==wxT("tbz2") || endExt==wxT("tbz") )  return tarbz;
if ( prevExt==wxT("tar") && ( endExt==wxT("bz2") || endExt==wxT("bz") ) )  return tarbz;

if ( endExt==wxT("gz") )    return gzip;                   // Since we've already tested for foo.tar.gz, it's safe to test for foo.gz
if ( endExt==wxT("bz2") )  return bzip;

return NotArchive;
}

void Archive::LoadHistory()    // Load the combobox history into the stringarray
{
config = wxConfigBase::Get();
if (config==NULL)  { wxLogError(_("Couldn't load configuration!?")); return; }

config->SetPath( wxT("/History/Archive/ArchiveFilePaths/") );

size_t count = config->GetNumberOfEntries();          // Count the entries
if ( !count ) { config->SetPath(wxT("/")); return; }

ArchiveHistory.Clear();

wxString item, key, histprefix(wxT("hist_"));  
for ( size_t n=0; n < count; ++n )                    // Create the key for every item, from hist_a to hist_('a'+count-1)
  {  key.Printf(wxT("%c"), wxT('a') + (wxChar)n );              // Make key hold a, b, c etc
    config->Read( histprefix+key, &item );            // Load the entry
    if ( item.IsEmpty() )   continue;
    ArchiveHistory.Add( item );                       // Store it in the array
  }

config->SetPath(wxT("/"));  
}

void Archive::LoadPrefs()      // Load the last-used values for checkboxes etc
{
config = wxConfigBase::Get();                   // Find the config data
if (config==NULL)  { wxLogError(_("Couldn't load configuration!?")); return; }

config->SetPath( wxT("/History/Archive/LastUsedValues/") );
wxString key, str; long lng;
                                                // Compress dialog
key = wxT("Slidervalue"); config->Read( key, &lng, (long)6 ); Slidervalue = (unsigned int)lng;
key = wxT("Compressionchoice"); config->Read( key, &Compressionchoice, 1 );
key = wxT("Recurse"); config->Read( key, &Recurse, 1 );
key = wxT("Force"); config->Read( key, &Force, 0 );
                                                // Decompress dialog
key = wxT("DecompressRecurse"); config->Read( key, &DecompressRecurse, 1 );
key = wxT("DecompressForce"); config->Read( key, &DecompressForce, 0 );
key = wxT("DecompressArchivesToo"); config->Read( key, &DecompressArchivesToo, 0 );

key = wxT("ArchiveExtractForce"); config->Read( key, &ArchiveExtractForce, 0 );  // Archive Extract dialog
                                                // CreateAppend dialogs
key = wxT("DerefSymlinks"); config->Read( key, &DerefSymlinks, 0 );
key = wxT("Verify"); config->Read( key, &Verify, 0 );
key = wxT("DeleteSource"); config->Read( key, &DeleteSource, 0 );
key = wxT("ArchiveCompress"); config->Read( key, &lng, (long)1 ); ArchiveCompress = (unsigned int)lng;

config->SetPath(wxT("/"));
}

void Archive::SaveHistory()  // Save the Archive filepath history
{
config = wxConfigBase::Get();
if (config==NULL)  { wxLogError(_("Couldn't load configuration!?")); return; }

config->DeleteGroup( wxT("/History/Archive/ArchiveFilePaths") );  // Delete current info, otherwise we'll end up with duplicates or worse
config->SetPath( wxT("/History/Archive/ArchiveFilePaths/") );

size_t count = wxMin( ArchiveHistory.GetCount(), MAX_COMMAND_HISTORY );  // Count the entries to be saved: not TOO many
if ( !count ) { config->SetPath(wxT("/")); return; }

wxString item, key, histprefix(wxT("hist_"));   // Create the key for every item, from hist_a to hist_('a'+count-1)
for ( size_t n=0;  n < count; ++n )             // The newest command is stored at the beginning of the array
  {  key.Printf(wxT("%c"), wxT('a') + (wxChar)n );       // Make key hold a, b, c etc
    config->Write( histprefix+key, ArchiveHistory[n] );  // Save the entry
  }

config->Flush();
config->SetPath(wxT("/"));  
}

void Archive::SavePrefs()    // Save the last-used values for checkboxes etc
{
config = wxConfigBase::Get();
if (config==NULL)  { wxLogError(_("Couldn't load configuration!?")); return; }

wxString path( wxT("/History/Archive/LastUsedValues/") );

                                                // Compress dialog
config->Write( path+wxT("Slidervalue"), (long)Slidervalue );
config->Write( path+wxT("Compressionchoice"), Compressionchoice );
config->Write( path+wxT("Recurse"), Recurse );
config->Write( path+wxT("Force"), Force );
                                                // Decompress dialog
config->Write( path+wxT("DecompressRecurse"), DecompressRecurse );
config->Write( path+wxT("DecompressForce"), DecompressForce );
config->Write( path+wxT("DecompressArchivesToo"), DecompressArchivesToo );
                                                // Archive Extract dialog
config->Write( path+wxT("ArchiveExtractForce"), ArchiveExtractForce );
                                                // CreateAppend dialogs
config->Write( path+wxT("DerefSymlinks"), DerefSymlinks );
config->Write( path+wxT("Verify"), Verify );
config->Write( path+wxT("DeleteSource"), DeleteSource );
config->Write( path+wxT("ArchiveCompress"), (long)ArchiveCompress );

config->Flush();
}


void Archive::UpdatePanes()  // Work out which dirs are most likely to have been altered, & update them
{
                // Updatedir may already have an entry, for the current dir
for ( size_t n=0; n < FilepathArray.Count(); ++n )  // Go thru the items that were in the listbox, adding their dirs to the array
  {  wxString path = FilepathArray[n];
    FileData fd( path );  
    if ( fd.IsRegularFile() || fd.IsSymlink() )  path = path.BeforeLast(wxFILE_SEP_PATH);  // We only want paths, not filepaths
    Updatedir.Add( path );                          // Store it
  }
if ( Updatedir.IsEmpty() ) return;

MyGenericDirCtrl::Clustering = true;
wxArrayInt IDs;
for ( size_t n=0; n < Updatedir.Count(); ++n )      // Go thru Updatedir, adding each item to the OnUpdateTrees() store
  MyFrame::mainframe->OnUpdateTrees( Updatedir[n], IDs );

MyFrame::mainframe->UpdateTrees();                  // Now flag actually to do the update
}


